/* $Id: detmon.c,v 1.11 2013-07-19 12:00:24 jtaylor Exp $
 *
 * This file is part of the irplib package
 * Copyright (C) 2002, 2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02111-1307 USA
 */

/*
 * $Author: jtaylor $
 * $Date: 2013-07-19 12:00:24 $
 * $Revision: 1.11 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*---------------------------------------------------------------------------
                                  Includes
 ---------------------------------------------------------------------------*/

#include "detmon_ronbias.h"
#include "detmon.h"

#include "irplib_ksigma_clip.h"
#include "irplib_hist.h"
#include "irplib_utils.h"

#include <math.h>
#include <string.h>
#include <assert.h>
#include <float.h>

/**@{*/
/*--------------------------------------------------------------------------*/

/*
 * @defgroup detmon        Detector monitoring functions
 */

/*--------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------
                                  Defines
 ---------------------------------------------------------------------------*/

/* Computes the square of an euclidean distance bet. 2 points */
#define pdist(x1,y1,x2,y2) (((x1-x2)*(x1-x2))+((y1-y2)*(y1-y2)))

#define cpl_drand() ((double)rand()/(double)RAND_MAX)

enum pixeltypes
{
    HOT = 0,
    DEAD = 1,
    NOISY = 2
};

enum stackingtypes
{
    MINMAX = 0,
    MEAN = 1,
    MEDIAN = 2,
    KSIGMA = 3
};

enum readouts
{
    HORIZONTAL = 1,
    VERTICAL = 2
};


static struct
{
    /* Inputs */
    const char             *method;
    const char             *pmethod;
    irplib_ronbias_method   method_bitmask;
    int                     prescan_llx;
    int                     prescan_lly;
    int                     prescan_urx;
    int                     prescan_ury;
    int                     overscan_llx;
    int                     overscan_lly;
    int                     overscan_urx;
    int                     overscan_ury;
    cpl_size                preoverscan_degree;
    int                     random_nsamples;
    int                     random_sizex;
    int                     random_sizey;
    int                     criteria;
    int                     ref_llx;
    int                     ref_lly;
    int                     ref_urx;
    int                     ref_ury;
    const char             *stacking_method;
    int                     stacking_ks_low;
    int                     stacking_ks_high;
    int                     stacking_ks_iter;
    int                     master_shift_x;
    int                     master_shift_y;
    int                     ron_llx;
    int                     ron_lly;
    int                     ron_urx;
    int                     ron_ury;
    int                     exts;
    int                     nb_extensions;
} detmon_ronbias_config;

#define NIR TRUE
#define OPT FALSE

/*---------------------------------------------------------------------------
  Private function prototypes
  ---------------------------------------------------------------------------*/

/* Functions for RON/Bias recipe, detmon_ronbias() */

static cpl_error_code
detmon_ronbias_retrieve_parlist(const char *,
                                const char *,
                                const cpl_parameterlist *,
                                cpl_boolean);

static cpl_error_code
detmon_ronbias_random(const cpl_imagelist *,
                      const cpl_image *, cpl_propertylist *);

static cpl_error_code
detmon_ronbias_histo(const cpl_imagelist *,
                     const cpl_image *, cpl_propertylist *);

static cpl_error_code
detmon_compute_bias_ron_and_stats_on_preoverscan(const cpl_imagelist *, /*const cpl_image *,*/
                           cpl_propertylist *, cpl_image **);

static cpl_error_code
detmon_compute_bias_stats_and_ron_in_region(const cpl_imagelist *,
                      const cpl_image *, cpl_propertylist *);

static cpl_image *
detmon_ronbias_master(const cpl_imagelist *,
                      cpl_mask **, cpl_mask **, cpl_mask **,
                      cpl_propertylist *);

static cpl_error_code
detmon_ronbias_save(const cpl_parameterlist *,
                    cpl_frameset *,
                    const char *,
                    const char *,
                    const char *,
                    const cpl_propertylist *,
                    const cpl_propertylist *,
                    const cpl_propertylist *,
                    const cpl_propertylist *,
                    const cpl_propertylist *,
                    const cpl_propertylist *,
                    const cpl_propertylist *,
                    const char *,
                    const cpl_image *,
                    const cpl_image *,
                    const cpl_mask *,
                    const cpl_mask *,
                    const cpl_mask *,
                    cpl_propertylist *,
                    const int,
                    const int,
                    cpl_frameset *,
                    int);

static int
detmon_ronbias_dfs_set_groups(cpl_frameset *, const char *);

static cpl_image *
detmon_build_synthetic_from_pre_overscan(const cpl_image * prescan,
                                         const cpl_image * overscan)
    CPL_ATTR_ALLOC;

static cpl_error_code
detmon_ronbias_dutycycl(const cpl_frameset *, cpl_propertylist *);



/* The following 2 functions are duplicated from cpl_det */



static cpl_bivector    *
irplib_bivector_gen_rect_poisson(const int *r,
                                 const int np,
                                 const int homog);

/* End of duplicated code */


cpl_error_code
detmon_ronbias_check_defaults(const cpl_frameset *, const int whichext);


/*         RONBIAS FILLING PARLIST        */



/*---------------------------------------------------------------------------*/

/*
 * @brief  Fill input parameters with default values
 * @param  parlist      parameters list
 * @param  recipe_name  recipe name
 * @param  pipeline_name pipeline name

 * @return CPL_ERROR_NONE on success.
 */

/*---------------------------------------------------------------------------*/

cpl_error_code
detmon_ronbias_fill_parlist_default(cpl_parameterlist * parlist,
                                    const char *recipe_name,
                                    const char *pipeline_name)
{
    const cpl_error_code error =
        detmon_ronbias_fill_parlist(parlist, recipe_name, pipeline_name,
                                    "ALL", /* --method */
                                    "NORM",/* --pmethod */
                                    1,     /* --preoverscan_degree */
                                    -1,    /* --random_nsamples */
                                    -1,    /* --random_sizex */
                                    -1,    /* --random_sizey */
                                    0,     /* --criteria     */
                                    -1,    /* --ref_llx     */
                                    -1,    /* --ref_lly */
                                    -1,    /* --ref_urx */
                                    -1,    /* --ref_ury */
                                    "MEAN",/* --stacking_method */
                                    3,     /* --stacking_ks_low */
                                    3,     /* --stacking_ks_high */
                                    5,     /* --stacking_ks_iter */
                                    0,     /* --master_shift_x */
                                    0,     /* --master_shift_y */
                                    -1,    /* --ron_llx */
                                    -1,    /* --ron_lly */
                                    -1,    /* --ron_urx */
                                    -1,    /* --ron_ury */
                                    0,    /* --exts */
                                    OPT);
    cpl_ensure_code(!error, error);

    return cpl_error_get_code();
}



/*---------------------------------------------------------------------------*/

/*
 * @brief  Fill input parameters with default values
 * @param  parlist      parameters list
 * @param  recipe_name  recipe name
 * @param  pipeline_name pipeline name
 * @param  method adopted method
 * @param  pmethod adopted pre-method
 * @param  preoverscan_degree degree used ti fit pre-overscan regions
 * @param  random_nsamples number of samples used for random computation
 * @param  random_sizex x size of rectangle area for random computation
 * @param  random_sizey x size of rectangle area for random computation
 * @param  criteria
 * @param  ref_llx  reference region lower left x
 * @param  ref_lly  reference region lower left y
 * @param  ref_urx  reference region upper right x
 * @param  ref_ury  reference region upper right y
 * @param  stacking_method  frame stacking method
 * @param  stacking_ks_low  kappa value to kappa sigma low intensity pixels
 * @param  stacking_ks_high  kappa value to kappa sigma high intensity pixels
 * @param  stacking_ks_iter  kappa value to kappa sigma number of iterations
 * @param  master_shift_x  x shift value applied to master
 * @param  master_shift_y  y shift value applied to master
 * @param  ron_llx  reference region lower left x to compute RON
 * @param  ron_lly  reference region lower left y to compute RON
 * @param  ron_urx  reference region upper right x to compute RON
 * @param  ron_ury  reference region upper right y to compute RON
 * @param  exts  image extension to be reduced
 * @param  opt_nir  switch to specify if in input are OPT or NIR data

 * @return CPL_ERROR_NONE on success.
 */

/*---------------------------------------------------------------------------*/

cpl_error_code
detmon_ronbias_fill_parlist(cpl_parameterlist * parlist,
                            const char *recipe_name,
                            const char *pipeline_name,
                            const char * method,
                            const char * pmethod,
                            const int preoverscan_degree,
                            const int random_nsamples,
                            const int random_sizex,
                            const int random_sizey,
                            const int criteria,
                            const int ref_llx,
                            const int ref_lly,
                            const int ref_urx,
                            const int ref_ury,
                            const char * stacking_method,
                            const int stacking_ks_low,
                            const int stacking_ks_high,
                            const int stacking_ks_iter,
                            const int master_shift_x,
                            const int master_shift_y,
                            const int ron_llx,
                            const int ron_lly,
                            const int ron_urx,
                            const int ron_ury,
                            const int exts,
                            cpl_boolean opt_nir)
{

    const char * meth_desc_opt =
        "Method to be used when computing bias. Methods appliable: "
        "<RANDOM | HISTO | PREOVERSCAN | REGION | ALL>. By default ALL "
        "methods are applied. More than a method can be chosen; in that "
        "case selected methods must be separated by a single space and put "
        "together between inverted commas (ex. --method=\"HISTO REGION\")."
        "\n RANDOM: Bias is computed as the mean value on a given number "
        "(--random.nsamples) of boxes (dimensions --random.sizex and "
        "--random.sizey) randomly taken  accross the detector.\n HISTO: "
        "An histogram of the pixels of the image is built.\n PREOVERSCAN: "
        "Mean, median and RMS values computed and designated areas. \n "
        "REGION: Mean, median and RMS values on reference region.";

    const char * meth_desc_nir =
        "Method to be used when computing bias. Methods appliable: "
        "<RANDOM | HISTO | REGION | ALL>. By default ALL "
        "methods are applied. More than a method can be chosen; in that "
        "case selected methods must be separated by a single space and put "
        "together between inverted commas (ex. --method=\"HISTO REGION\")."
        "\n RANDOM: Bias is computed as the mean value on a given number "
        "(--random.nsamples) of boxes (dimensions --random.sizex and "
        "--random.sizey) randomly taken  accross the detector.\n HISTO: "
        "An histogram of the pixels of the image is built.\n "
        "REGION: Mean, median and RMS values on reference region.";

    const char * method_desc = opt_nir == OPT ? meth_desc_opt : meth_desc_nir;

    const cpl_error_code error =
        detmon_fill_parlist(parlist, recipe_name, pipeline_name, 22,
                            "method",
                            method_desc,
                            "CPL_TYPE_STRING", method,

                            "pmethod",
                            "Pre-method for RANDOM, HISTO and REGION."
                            "Difference raw frames or not",
                            "CPL_TYPE_STRING", pmethod,

                            "preoverscan.degree",
                            "Degree used for pre-overscan method",
                            "CPL_TYPE_INT", preoverscan_degree,

                            "random.nsamples",
                            "Number of samples",
                            "CPL_TYPE_INT", random_nsamples,

                            "random.sizex",
                            "X size of the boxes",
                            "CPL_TYPE_INT", random_sizex,

                            "random.sizey",
                            "Y size of the boxes",
                            "CPL_TYPE_INT", random_sizey,

                            "criteria",
                            "Criteria",
                            "CPL_TYPE_INT", criteria,

                            "ref.llx",
                            "x coordinate of the lower-left point "
                                "of the reference region of the frame",
                            "CPL_TYPE_INT", ref_llx,

                            "ref.lly",
                            "y coordinate of the lower-left point "
                                "of the reference region of the frame",
                            "CPL_TYPE_INT", ref_lly,

                            "ref.urx",
                            "x coordinate of the upper-right point "
                                "of the reference region of the frame",
                            "CPL_TYPE_INT", ref_urx,

                            "ref.ury",
                            "y coordinate of the upper-right point "
                                "of the reference region of the frame",
                            "CPL_TYPE_INT", ref_ury,

                            "stacking.method",
                            "Method to be used when stacking the master. Posible values < MINMAX | MEAN | MEDIAN | KSIGMA >",
                            "CPL_TYPE_STRING", stacking_method,

                            "stacking.ks.low",
                            "Low threshold for kappa-sigma clipping",
                            "CPL_TYPE_INT", stacking_ks_low,

                            "stacking.ks.high",
                            "High threshold for kappa-sigma clipping",
                            "CPL_TYPE_INT", stacking_ks_high,

                            "stacking.ks.iter",
                            "Nb of iterations for kappa-sigma clipping",
                            "CPL_TYPE_INT", stacking_ks_iter,

                            "master.shift.x",
                            "Master shift X",
                            "CPL_TYPE_INT", master_shift_x,

                            "master.shift.y",
                            "Master shift Y",
                            "CPL_TYPE_INT", master_shift_y,

                            "ron.llx",
                            "x coordinate of the lower-left point "
                                "of the RON frame",
                            "CPL_TYPE_INT", ron_llx,

                            "ron.lly",
                            "y coordinate of the lower-left point "
                                "of the RON frame",
                            "CPL_TYPE_INT", ron_lly,

                            "ron.urx",
                            "x coordinate of the upper-right point "
                                "of the RON frame",
                            "CPL_TYPE_INT", ron_urx,

                            "ron.ury",
                            "y coordinate of the upper-right point "
                                "of the RON frame", "CPL_TYPE_INT", ron_ury,

                            "exts",
                            "Activate the multi-exts option",
                            "CPL_TYPE_INT", exts);


    cpl_ensure_code(!error, error);

    return cpl_error_get_code();
}



#define DETMON_MAX_PARAM 19

/*---------------------------------------------------------------------------*/

/*
 * @brief  Retrieve input parameters
 * @param  pipeline_name        Input image
 * @param  recipe_name          Input image
 * @param  parlist              Shift to apply on the x-axis
 * @param  opt_nir  switch to specify if in input are OPT or NIR data
 * @return CPL_ERROR_NONE on success.
 */

/*---------------------------------------------------------------------------*/
static cpl_error_code
detmon_ronbias_retrieve_parlist(const char *pipeline_name,
                                const char *recipe_name,
                                const cpl_parameterlist * parlist,
                                cpl_boolean opt_nir)
{
    char                   *par_name;
    const cpl_parameter    *par;

    char                    m1[DETMON_MAX_PARAM+1] = "";
    char                    m2[DETMON_MAX_PARAM+1] = "";
    char                    m3[DETMON_MAX_PARAM+1] = "";

    /* --method */
    par_name = cpl_sprintf("%s.%s.method", pipeline_name, recipe_name);
    assert(par_name != NULL);
    par = cpl_parameterlist_find_const(parlist, par_name);
    detmon_ronbias_config.method = cpl_parameter_get_string(par);
    cpl_free(par_name);

    detmon_ronbias_config.method_bitmask = 0;

    sscanf(detmon_ronbias_config.method,
           "%" CPL_STRINGIFY(DETMON_MAX_PARAM) "s "
           "%" CPL_STRINGIFY(DETMON_MAX_PARAM) "s "
           "%" CPL_STRINGIFY(DETMON_MAX_PARAM) "s", m1, m2, m3);

    if(!strcmp(m1, "RANDOM") || !strcmp(m2, "RANDOM")
       || !strcmp(m3, "RANDOM"))
        detmon_ronbias_config.method_bitmask += RANDOM;

    if(!strcmp(m1, "HISTO") || !strcmp(m2, "HISTO") || !strcmp(m3, "HISTO"))
        detmon_ronbias_config.method_bitmask += HISTO;

    if(!strcmp(m1, "PREOVERSCAN") || !strcmp(m2, "PREOVERSCAN")
       || !strcmp(m3, "PREOVERSCAN")) {
        if (opt_nir == NIR) {
            /* Just in case some advance user reads himself in the code and
               tries to trick the interface providing an option no contained
               in the man-page */
            cpl_msg_warning(cpl_func, "PREOVERSCAN is not appliable for NIR");
        } else {
            detmon_ronbias_config.method_bitmask += PREOVERSCAN;
        }
    }
    if(!strcmp(m1, "REGION") || !strcmp(m2, "REGION")
       || !strcmp(m3, "REGION"))
        detmon_ronbias_config.method_bitmask += REGION;

    if(!strcmp(m1, "ALL")) {
        if (opt_nir == OPT) {
            detmon_ronbias_config.method_bitmask =
                RANDOM | HISTO | PREOVERSCAN | REGION;
        } else {
            detmon_ronbias_config.method_bitmask =
                RANDOM | HISTO | REGION;
        }
    }

    /* --pmethod */
    par_name = cpl_sprintf("%s.%s.pmethod", pipeline_name, recipe_name);
    assert(par_name != NULL);
    par = cpl_parameterlist_find_const(parlist, par_name);
    detmon_ronbias_config.pmethod = cpl_parameter_get_string(par);
    cpl_free(par_name);

    /* --preoverscan.degree */
    detmon_ronbias_config.preoverscan_degree =
        detmon_retrieve_par_int("preoverscan.degree", pipeline_name,
                                recipe_name, parlist);

    /* --nsamples */
    detmon_ronbias_config.random_nsamples =
        detmon_retrieve_par_int("random.nsamples", pipeline_name,
                                recipe_name, parlist);

    /* --sizex */
    detmon_ronbias_config.random_sizex =
        detmon_retrieve_par_int("random.sizex", pipeline_name,
                                recipe_name, parlist);

    /* --sizey */
    detmon_ronbias_config.random_sizey =
        detmon_retrieve_par_int("random.sizey", pipeline_name,
                                recipe_name, parlist);

    /* --criteria */
    detmon_ronbias_config.criteria =
        detmon_retrieve_par_int("criteria", pipeline_name, recipe_name,
                                parlist);

    /* --ref.llx */
    detmon_ronbias_config.ref_llx =
        detmon_retrieve_par_int("ref.llx", pipeline_name, recipe_name,
                                parlist);
    /* --ref.lly */
    detmon_ronbias_config.ref_lly =
        detmon_retrieve_par_int("ref.lly", pipeline_name, recipe_name,
                                parlist);
    /* --ref.urx */
    detmon_ronbias_config.ref_urx =
        detmon_retrieve_par_int("ref.urx", pipeline_name, recipe_name,
                                parlist);
    /* --ref.ury */
    detmon_ronbias_config.ref_ury =
        detmon_retrieve_par_int("ref.ury", pipeline_name, recipe_name,
                                parlist);

    /* --stacking.method */
    par_name =
        cpl_sprintf("%s.%s.stacking.method", pipeline_name, recipe_name);
    assert(par_name != NULL);
    par = cpl_parameterlist_find_const(parlist, par_name);
    detmon_ronbias_config.stacking_method = cpl_parameter_get_string(par);
    cpl_free(par_name);

    /* --stacking.ks.low */
    detmon_ronbias_config.stacking_ks_low =
        detmon_retrieve_par_int("stacking.ks.low", pipeline_name,
                                recipe_name, parlist);
    /* --stacking.ks.high */
    detmon_ronbias_config.stacking_ks_high =
        detmon_retrieve_par_int("stacking.ks.high", pipeline_name,
                                recipe_name, parlist);
    /* --stacking.ks.iter */
    detmon_ronbias_config.stacking_ks_iter =
        detmon_retrieve_par_int("stacking.ks.iter", pipeline_name,
                                recipe_name, parlist);
    /* --master.shift.x */
    detmon_ronbias_config.master_shift_x =
        detmon_retrieve_par_int("master.shift.x", pipeline_name,
                                recipe_name, parlist);
    /* --master.shift.y */
    detmon_ronbias_config.master_shift_y =
        detmon_retrieve_par_int("master.shift.y", pipeline_name,
                                recipe_name, parlist);
    /* --ron.llx */
    detmon_ronbias_config.ron_llx =
        detmon_retrieve_par_int("ron.llx", pipeline_name, recipe_name,
                                parlist);
    /* --ron.lly */
    detmon_ronbias_config.ron_lly =
        detmon_retrieve_par_int("ron.lly", pipeline_name, recipe_name,
                                parlist);
    /* --ron.urx */
    detmon_ronbias_config.ron_urx =
        detmon_retrieve_par_int("ron.urx", pipeline_name, recipe_name,
                                parlist);
    /* --ron.ury */
    detmon_ronbias_config.ron_ury =
        detmon_retrieve_par_int("ron.ury", pipeline_name, recipe_name,
                                parlist);
    /* --exts */
    detmon_ronbias_config.exts =
        detmon_retrieve_par_int("exts", pipeline_name, recipe_name,
                                parlist);

    if(cpl_error_get_code()) {
        cpl_msg_error(cpl_func, "Failed to retrieve the input parameters");
        cpl_ensure_code(0, CPL_ERROR_DATA_NOT_FOUND);
    }


    return CPL_ERROR_NONE;
}

/*---------------------------------------------------------------------------*/

/*
 * @brief  Check parameter defaults
 * @param  set        Input set of frames
 * @param  whichext   extension to be reduced
 * @return CPL_ERROR_NONE on success.
 */

/*---------------------------------------------------------------------------*/
cpl_error_code
detmon_ronbias_check_defaults(const cpl_frameset * set,
                              const int whichext)
{
    const cpl_frame  * fr        = cpl_frameset_get_position_const(set, 0);

    cpl_propertylist * plist     =
        cpl_propertylist_load(cpl_frame_get_filename(fr), whichext);

    const int naxis1 = cpl_propertylist_get_int(plist, "NAXIS1");
    const int naxis2 = cpl_propertylist_get_int(plist, "NAXIS2");

    if(detmon_ronbias_config.method_bitmask & PREOVERSCAN)
    {
        const int nx = cpl_propertylist_get_int(plist, "ESO DET OUT1 NX");
        const int ny = cpl_propertylist_get_int(plist, "ESO DET OUT1 NY");

        int prscsize;
        int ovscsize;

        if (naxis1 != nx)
        {
            prscsize =
                cpl_propertylist_get_int(plist, "ESO DET OUT1 PRSCX");
            ovscsize =
                cpl_propertylist_get_int(plist, "ESO DET OUT1 OVSCX");

            cpl_error_ensure(cpl_error_get_code() == CPL_ERROR_NONE, cpl_error_get_code(), goto cleanup,"error");

            detmon_ronbias_config.prescan_llx  = 1;
            detmon_ronbias_config.prescan_lly  = 1;
            detmon_ronbias_config.prescan_urx  = prscsize;
            detmon_ronbias_config.prescan_ury  = naxis2;
            detmon_ronbias_config.overscan_llx = naxis1 - ovscsize;
            detmon_ronbias_config.overscan_lly = 1;
            detmon_ronbias_config.overscan_urx = naxis1;
            detmon_ronbias_config.overscan_ury = naxis2;
        } else if (naxis2 != ny)
        {
            prscsize =
                cpl_propertylist_get_int(plist, "ESO DET OUT1 PRSCY");
            ovscsize =
                cpl_propertylist_get_int(plist, "ESO DET OUT1 OVSCY");
            cpl_error_ensure(cpl_error_get_code() == CPL_ERROR_NONE, cpl_error_get_code(), goto cleanup,"error");

            detmon_ronbias_config.prescan_llx  = 1;
            detmon_ronbias_config.prescan_lly  = 1;
            detmon_ronbias_config.prescan_urx  = naxis1;
            detmon_ronbias_config.prescan_ury  = prscsize;
            detmon_ronbias_config.overscan_llx = 1;
            detmon_ronbias_config.overscan_lly = naxis2 - ovscsize;
            detmon_ronbias_config.overscan_urx = naxis1;
            detmon_ronbias_config.overscan_ury = naxis2;
        } else
        {
            cpl_msg_error(cpl_func,
                          "No PREOVERSCAN areas found");
            cpl_error_set(cpl_func, CPL_ERROR_NULL_INPUT);
            goto cleanup;
        }
    }

    if(detmon_ronbias_config.ref_llx == -1)
        detmon_ronbias_config.ref_llx = naxis1 / 8;
    if(detmon_ronbias_config.ref_lly == -1)
        detmon_ronbias_config.ref_lly = naxis2 / 8;
    if(detmon_ronbias_config.ref_urx == -1)
        detmon_ronbias_config.ref_urx = naxis1 * 7 / 8;
    if(detmon_ronbias_config.ref_ury == -1)
        detmon_ronbias_config.ref_ury = naxis2 * 7 / 8;

    if(detmon_ronbias_config.ron_llx == -1)
        detmon_ronbias_config.ron_llx = 1;
    if(detmon_ronbias_config.ron_lly == -1)
        detmon_ronbias_config.ron_lly = 1;
    if(detmon_ronbias_config.ron_urx == -1)
        detmon_ronbias_config.ron_urx = naxis1;
    if(detmon_ronbias_config.ron_ury == -1)
        detmon_ronbias_config.ron_ury = naxis2;

cleanup:
    cpl_propertylist_delete(plist);
    return cpl_error_get_code();
}


/*---------------------------------------------------------------------------*/

/*
 * @brief  save a ronbias results on a FITS file
 * @param  parlist        input parameters
 * @param  tag            input frame tag
 * @param  recipe_name    input recipe name
 * @param  pipeline_name  input pipeline id
 * @param  pafregexp      regular expression to remove from FITS header
 *
 * @param  pro_master     pro catg master frame
 * @param  pro_xstr       pro catg xstructure
 * @param  pro_ystr       pro catg ystructure
 * @param  pro_synth      pro catg synthetic frame
 * @param  pro_bpmhot     pro catg hot pix BPM
 * @param  pro_pmpcold    pro catg cold pix BPM
 * @param  pro_ppmdev     pro catg deviant pix BPM
 *
 * @param  package        package name
 * @param  compare        function used for comparison
 * @param  opt_nir        switch to specify if in input are OPT or NIR data

 * @return CPL_ERROR_NONE on success.
 */

/*---------------------------------------------------------------------------*/
cpl_error_code
detmon_ronbias(cpl_frameset * frameset,
               const cpl_parameterlist * parlist,
               const char *tag,
               const char *recipe_name,
               const char *pipeline_name,
               const char *pafregexp,
               const cpl_propertylist * pro_master,
               const cpl_propertylist * pro_xstr, /* Unsupported*/
               const cpl_propertylist * pro_ystr, /* Unsupported*/
               const cpl_propertylist * pro_synth,
               const cpl_propertylist * pro_bpmhot,
               const cpl_propertylist * pro_bpmcold,
               const cpl_propertylist * pro_bpmdev,
               const char *package,
               int (*compare) (const cpl_frame *, const cpl_frame *),
               cpl_boolean opt_nir)
{

    cpl_size                     nsets;
    int                     i;

    cpl_size                    * selection  = NULL;
    cpl_frameset           * cur_fset   = NULL;
    cpl_propertylist       * qclist     = NULL;
    cpl_image              * synthetic  = NULL;
    cpl_image              * masterbias = NULL;
    cpl_imagelist          * rawbiases  = NULL;
    cpl_mask               * bpmhot     = NULL;
    cpl_mask               * bpmcold    = NULL;
    cpl_mask               * bpmdev     = NULL;

    /* Test entries */
    cpl_ensure_code(frameset        != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(parlist         != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(tag             != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(recipe_name     != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(pipeline_name   != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(pro_master      != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(pro_bpmhot      != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(pro_bpmcold     != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(pro_bpmdev      != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(package         != NULL, CPL_ERROR_NULL_INPUT);

    if(detmon_ronbias_dfs_set_groups(frameset, tag)) {
        cpl_msg_error(cpl_func, "Cannot identify RAW and CALIB frames");
    }

    /*
     * First of all test the entries.
     * See if the selected method(s) is/are appliable.
     * See if necessary parameters for those selected have been provided.
     */

    /*    clreturn_if(detmon_ronbias_test_entries());
    */
    /*
     * This function reads all inputs parameters from parlist
     * and stores them in a global variable detmon_ronbias_config.
     * Similar to detmon_lg_retrieve_parlist(). See detmon.c
     */
    detmon_ronbias_retrieve_parlist(pipeline_name,
                                    recipe_name, parlist, opt_nir);

    /* Extra input check for PREOVERSCAN */
    if(detmon_ronbias_config.method_bitmask & PREOVERSCAN)
        cpl_ensure_code(pro_synth       != NULL, CPL_ERROR_NULL_INPUT);


    /* Labelise all input frames */
    if(compare == NULL)
        nsets = 1;
    else {
        cpl_msg_info(cpl_func, "Identify the different settings");
        selection = cpl_frameset_labelise(frameset, compare, &nsets);
        if(selection == NULL)
            cpl_msg_error(cpl_func, "Cannot labelise input frames");
    }

    /* Extract settings and reduce each of them */
    for(i = 0; i < nsets; i++) {
        int  j;
        int first_ext = 0;
        int last_ext  = 1;

        detmon_ronbias_config.nb_extensions = 1;

        /* Reduce data set nb i */
        cpl_msg_info(cpl_func, "Reduce data set nb %d out of %" CPL_SIZE_FORMAT "",
                     i + 1, nsets);

        cur_fset = nsets == 1 ?
            cpl_frameset_duplicate(frameset) :
            cpl_frameset_extract(frameset, selection, i);
        skip_if(cur_fset == NULL);

        if(detmon_ronbias_config.exts > 0) {
            first_ext = detmon_ronbias_config.exts;
            last_ext  = first_ext + 1;
        } else if(detmon_ronbias_config.exts < 0) {
            const cpl_frame        *cur_frame =
                cpl_frameset_get_position_const(cur_fset, 0);
            /* Get the nb of extensions */
            detmon_ronbias_config.nb_extensions =
                cpl_frame_get_nextensions(cur_frame);
            first_ext = 1;
            last_ext = detmon_ronbias_config.nb_extensions + 1;
        }

        if (last_ext - first_ext > 1) {
            skip_if(detmon_ronbias_save(parlist, frameset,
                                        recipe_name,
                                        pipeline_name, pafregexp,
                                        pro_master, pro_xstr,
                                        pro_ystr, pro_synth,
                                        pro_bpmhot,
                                        pro_bpmcold, pro_bpmdev,
                                        package, NULL, NULL, NULL,
                                        NULL, NULL, NULL,
                                        0, 0, cur_fset, 0));
        }

        for(j = first_ext; j < last_ext; j++) {
            int whichext;

            qclist = cpl_propertylist_new();

            rawbiases
                = cpl_imagelist_load_frameset(cur_fset,
                                              CPL_TYPE_FLOAT, 1, j);
            skip_if(rawbiases == NULL);

            skip_if(detmon_ronbias_check_defaults(cur_fset, j));

            skip_if(detmon_ronbias_dutycycl(cur_fset, qclist));

            masterbias = detmon_ronbias_master(rawbiases,
                                               &bpmhot, &bpmcold,
                                               &bpmdev, qclist);
            skip_if(masterbias == NULL);

            /*
             * Following, a function corresponding each of the
             * possible methods is to be found.
             */

            if(detmon_ronbias_config.method_bitmask & RANDOM) {
                skip_if(detmon_ronbias_random(rawbiases, masterbias,
                                              qclist));
            }

            if(detmon_ronbias_config.method_bitmask & HISTO) {
                skip_if(detmon_ronbias_histo(rawbiases, masterbias,
                                             qclist));
            }

            if(detmon_ronbias_config.method_bitmask & PREOVERSCAN) {
                skip_if(detmon_compute_bias_ron_and_stats_on_preoverscan(rawbiases,
                                                   /*masterbias,*/
                                                   qclist, &synthetic));
            }

            if(detmon_ronbias_config.method_bitmask & REGION) {
                skip_if(detmon_compute_bias_stats_and_ron_in_region(rawbiases,
                                masterbias,qclist));
            }

            /*
             * This function takes the QC list where all the results of the
             * methods applied are stored, and compares them.
             * No action defined yet if comparison reveals important differences.
             */
#if 0
            detmon_ronbias_check(qclist);
#endif

            /* Definition of the extension of the output where to save
               the products. If input are multiextension but only
               computation on a single extension is required, it is 0 */
            whichext = first_ext > 1 ? 0 : j;

            skip_if(detmon_ronbias_save(parlist, frameset,
                                        recipe_name,
                                        pipeline_name, pafregexp,
                                        pro_master, pro_xstr,
                                        pro_ystr, pro_synth,
                                        pro_bpmhot,
                                        pro_bpmcold, pro_bpmdev,
                                        package, masterbias, synthetic,
                                        bpmhot, bpmcold, bpmdev,
                                        qclist, 0, 0, cur_fset,
                                        whichext));

            cpl_image_delete(synthetic);
            cpl_image_delete(masterbias);
            cpl_mask_delete(bpmhot);
            cpl_mask_delete(bpmcold);
            cpl_mask_delete(bpmdev);
            cpl_imagelist_delete(rawbiases);
            cpl_propertylist_delete(qclist);

            qclist     = NULL;
            rawbiases  = NULL;
            masterbias = NULL;
            bpmhot     = NULL;
            bpmcold    = NULL;
            bpmdev     = NULL;
            synthetic  = NULL;
        } /* for each extension */

        cpl_frameset_delete(cur_fset);
        cur_fset = NULL;

    } /* for each setting */

    end_skip;

    cpl_free(selection);

    cpl_frameset_delete(cur_fset);

    cpl_image_delete(synthetic);
    cpl_image_delete(masterbias);
    cpl_mask_delete(bpmhot);
    cpl_mask_delete(bpmcold);
    cpl_mask_delete(bpmdev);
    cpl_imagelist_delete(rawbiases);
    cpl_propertylist_delete(qclist);

    return cpl_error_get_code();
}

/*---------------------------------------------------------------------------*/
/*
 * @brief  Computes bias RON using random sampling method
 * @param rawbiases   input image list
 * @param masterbias  input master bias
 * @param qclist      input/output propertylist with QC parameters
 * @return CPL_ERROR_NONE on success.

 */
/*---------------------------------------------------------------------------*/
static cpl_error_code
detmon_ronbias_random(const cpl_imagelist * rawbiases,
                      const cpl_image     * masterbias,
                      cpl_propertylist    * qclist)
{
    int                     nraws = cpl_imagelist_get_size(rawbiases);
    int                     i;
    double                  bias = DBL_MAX;  /* Avoid (false) uninit warning */
    double                  bias_error;

    double                  ron_error;
    double                 *ron =
        (double *) cpl_malloc(sizeof(double) * nraws);

    cpl_vector             *v;
    double stdev = 0;
    cpl_error_code error = CPL_ERROR_NONE;

    /* As we are applying to diff frames instead of raw frames,
       there is one less to compute on */
    if(!strcmp(detmon_ronbias_config.pmethod, "DIF"))
    {
        nraws--;
        /* As we are applying to diff frames instead of raw frames,
           there is one less to compute on */
        for(i = 0; i < nraws; i++)
        {
            const cpl_image        *c1_raw =
                cpl_imagelist_get_const(rawbiases, i);
            const cpl_image        *c2_raw =
                cpl_imagelist_get_const(rawbiases, i + 1);
            /*FIXME: See if const modifier is necessary */
            const cpl_image        *c_raw = cpl_image_subtract_create(c1_raw,
                                                                      c2_raw);
            error = cpl_flux_get_noise_window(c_raw, NULL,
                                              detmon_ronbias_config.random_sizex / 2,
                                              detmon_ronbias_config.random_nsamples,
                                              ron + i, &ron_error);
            cpl_image_delete((cpl_image*)c_raw);
            if (error != CPL_ERROR_NONE)
            {
                break;
            }
        }
    } else
    {
        for(i = 0; i < nraws; i++)
        {
            const cpl_image *c_raw = cpl_imagelist_get_const(rawbiases, i);
            skip_if(cpl_flux_get_noise_window(c_raw, NULL,
                                              detmon_ronbias_config.random_sizex / 2,
                                              detmon_ronbias_config.random_nsamples,
                                              ron + i, &ron_error));
        }
    }

    /*FIXME: Calls to noise_window could be out from if() and
      coded only once */
    if (error == CPL_ERROR_NONE)
    {
        irplib_flux_get_bias_window(masterbias, NULL,
                                    detmon_ronbias_config.random_sizex / 2,
                                    detmon_ronbias_config.random_nsamples,
                                    &bias, &bias_error);

        v = cpl_vector_wrap(nraws, ron);
        stdev = cpl_vector_get_median_const(v);
        cpl_vector_unwrap(v);


        skip_if(cpl_propertylist_append_double(qclist,DETMON_QC_BIAS_RANDOM_VAL, bias));
        skip_if(cpl_propertylist_set_comment(qclist,DETMON_QC_BIAS_RANDOM_VAL,
                                             DETMON_QC_BIAS_RANDOM_VAL_C));

        skip_if(cpl_propertylist_append_double(qclist,DETMON_QC_BIAS_RANDOM_RON, stdev));
        skip_if(cpl_propertylist_set_comment(qclist,DETMON_QC_BIAS_RANDOM_RON,
                                             DETMON_QC_BIAS_RANDOM_RON_C));
    }

    irplib_flux_get_bias_window(masterbias, NULL,
                                detmon_ronbias_config.random_sizex / 2,
                                detmon_ronbias_config.random_nsamples,
                                &bias, &bias_error);

    v = cpl_vector_wrap(nraws, ron);
    if (v)
    {
        stdev = cpl_vector_get_median_const(v);
        cpl_vector_unwrap(v);
    }

    error = cpl_propertylist_append_double(qclist,DETMON_QC_BIAS_RANDOM_VAL, bias);
    error = cpl_propertylist_set_comment(qclist,DETMON_QC_BIAS_RANDOM_VAL,
                                         DETMON_QC_BIAS_RANDOM_VAL_C);


    error = cpl_propertylist_append_double(qclist,DETMON_QC_BIAS_RANDOM_RON, stdev);
    error = cpl_propertylist_set_comment(qclist,DETMON_QC_BIAS_RANDOM_RON,
                                         DETMON_QC_BIAS_RANDOM_RON_C);

    end_skip;
    if (ron)
        cpl_free(ron);
    return cpl_error_get_code();
}
/*
 * @brief  Computes bias RON using histogram method
 * @param rawbiases   input image list
 * @param masterbias  input master bias
 * @param qclist      input/output propertylist with QC parameters
 * @return CPL_ERROR_NONE on success.

 */
static cpl_error_code
detmon_ronbias_histo(const cpl_imagelist * rawbiases,
                     const cpl_image * masterbias,
                     cpl_propertylist * qclist)
{
    int                     nraws = cpl_imagelist_get_size(rawbiases);
    int                     i;

    double mbias = DBL_MAX;
    double mfwhm = DBL_MAX;
    double mmax = DBL_MAX;

    cpl_vector * fwhms;
    cpl_vector * maxs;

    double mean_fwhm = DBL_MAX;

    if(!strcmp(detmon_ronbias_config.pmethod, "DIF")) nraws--;

    fwhms = cpl_vector_new(nraws);
    maxs = cpl_vector_new(nraws);

    for(i = 0; i < nraws; i++) {
        /*FIXME: See if it is necessary to have const */
        const cpl_image * c_raw;
        double bias = DBL_MAX;
        double fwhm = DBL_MAX;
        double max = DBL_MAX;

        if(strcmp(detmon_ronbias_config.pmethod, "DIF")) {
            c_raw = cpl_imagelist_get_const(rawbiases, i);
        } else {
            const cpl_image        *c1_raw = cpl_imagelist_get_const(rawbiases, i);
            const cpl_image        * c2_raw = cpl_imagelist_get_const(rawbiases, i+1);
            c_raw = cpl_image_subtract_create(c1_raw, c2_raw);
        }

        skip_if(detmon_ronbias_histo_reduce(c_raw, &bias, &fwhm, &max));

        skip_if(bias == DBL_MAX || fwhm == DBL_MAX || max == DBL_MAX);

        if(!strcmp(detmon_ronbias_config.pmethod, "DIF"))
            cpl_image_delete((cpl_image *)c_raw);

        skip_if(cpl_vector_set(maxs, i, max));
        skip_if(cpl_vector_set(fwhms, i, fwhm));

        /* FIXME: Add properly a hist-saving in debug-mode */
    }

    skip_if(cpl_vector_divide_scalar(fwhms, CPL_MATH_FWHM_SIG));

    detmon_ronbias_histo_reduce(masterbias, &mbias, &mfwhm, &mmax);

    skip_if(mbias == DBL_MAX || mfwhm == DBL_MAX || mmax == DBL_MAX);

    skip_if(cpl_propertylist_append_double(qclist,DETMON_QC_BIAS_HISTO_VAL,
                                           mbias));
    skip_if(cpl_propertylist_set_comment(qclist,DETMON_QC_BIAS_HISTO_VAL,
                                         DETMON_QC_BIAS_HISTO_VAL_C));
    mean_fwhm = cpl_vector_get_mean(fwhms);

    skip_if(mean_fwhm == DBL_MAX);
    skip_if(cpl_error_get_code());

    skip_if(cpl_propertylist_append_double(qclist,DETMON_QC_BIAS_HISTO_RON,
                                           mean_fwhm));
    skip_if(cpl_propertylist_set_comment(qclist,DETMON_QC_BIAS_HISTO_RON,
                                         DETMON_QC_BIAS_HISTO_RON_C));

    end_skip;

    cpl_vector_delete(fwhms);
    cpl_vector_delete(maxs);

    return cpl_error_get_code();
}

/*---------------------------------------------------------------------------*/

/*
 * @brief  Computes histogram on input image and get bias, fwhm and max
 * @param c_raw  Input image
 * @param bias   output bias of histogram
 * @param fwhm   output fwhm of histogram
 * @param max    output max of histogram
 * @return CPL_ERROR_NONE on success.

 */

/*---------------------------------------------------------------------------*/
cpl_error_code
detmon_ronbias_histo_reduce(const cpl_image * c_raw,
                            double * bias,
                            double * fwhm,
                            double * max)
{
    unsigned long           uj;
    irplib_hist            *hist;
    unsigned long           maxwhere = 0;

    cpl_image * dupi;
    unsigned long           x1a = 1;
    unsigned long           x2a = 1;

    double                  x1 = 0;
    double                  x2 = 0;

    double                  maxwhere_interp;
    double                  max_interp;
    double a, b, c;
    cpl_matrix * coeffs =cpl_matrix_new(3, 3);
    cpl_matrix * rhs    =cpl_matrix_new(3, 1);
    int p, q;
    cpl_matrix * result = NULL;

    dupi = cpl_image_duplicate(c_raw);

    /* FIXME: Still to decide if it is necessary to remove bad pixels
       in advance or not*/

    hist = irplib_hist_new();
    irplib_hist_fill(hist, dupi);

    cpl_image_delete(dupi);

    irplib_hist_get_max(hist, &maxwhere);

    for( p = 0; p< 3; p++){
        unsigned long bi = irplib_hist_get_value(hist, maxwhere-1+p);
        cpl_matrix_set(rhs, p, 0, bi);
        for( q= 0; q< 3; q++) {
            cpl_matrix_set(coeffs, p,q,pow((maxwhere-1+p),q));
        }
    }

    result = cpl_matrix_solve(coeffs, rhs);

    a = cpl_matrix_get(result, 2, 0);
    b = cpl_matrix_get(result, 1, 0);
    c = cpl_matrix_get(result, 0, 0);

    maxwhere_interp = -0.5 * b / (2 * a);
    max_interp = -1 * b * b / (4 * a) + c;

    cpl_matrix_delete(coeffs);
    cpl_matrix_delete(rhs);
    cpl_matrix_delete(result);

    /* Look for the points of half-maximum */
    for(uj = 0; uj < maxwhere; uj++) {
        if(irplib_hist_get_value(hist, uj)     <= max_interp / 2 &&
           irplib_hist_get_value(hist, uj + 1) >  max_interp / 2) {
            x1a = uj;
        }
    }
    for(uj = maxwhere; uj < irplib_hist_get_nbins(hist)-1; uj++) {
        if(irplib_hist_get_value(hist, uj)     >= max_interp / 2 &&
           irplib_hist_get_value(hist, uj + 1) <  max_interp / 2) {
            x2a = uj;
        }
    }

    x1 = (max_interp / 2 - irplib_hist_get_value(hist, x1a)) /
        (irplib_hist_get_value(hist, x1a + 1) -
         irplib_hist_get_value(hist, x1a)) + x1a;
    x2 = (max_interp / 2 - irplib_hist_get_value(hist, x2a)) /
        (irplib_hist_get_value(hist, x2a + 1) -
         irplib_hist_get_value(hist, x2a)) + x2a;

    *fwhm = (x2 - x1) * irplib_hist_get_bin_size(hist);

    *max = max_interp;

    *bias =  maxwhere_interp * irplib_hist_get_bin_size(hist) +
        irplib_hist_get_start(hist);

    irplib_hist_delete(hist);

    return cpl_error_get_code();
}
/*---------------------------------------------------------------------------*/

/*
 * @brief  Computes RON and statistics on pre/overscan bias regions
 * @param rawbiases        Input image list
 * @param qclist           input/output propertylist with QC parameters
 * @param synthetic        input synthetic image
 * @return CPL_ERROR_NONE on success.

 */

/*---------------------------------------------------------------------------*/
static cpl_error_code
detmon_compute_bias_ron_and_stats_on_preoverscan(const cpl_imagelist * rawbiases,
                           cpl_propertylist * qclist,
                           cpl_image ** synthetic)
{
    int                     i;
    int                     nx, ny;
    int                     nraws;

    cpl_vector             *meanspre;
    cpl_vector             *medspre;
    cpl_vector             *rmsspre;
    cpl_vector             *meansover;
    cpl_vector             *medsover;
    cpl_vector             *rmssover;

    cpl_error_code          error;

    nraws = cpl_imagelist_get_size(rawbiases);
    cpl_ensure_code(nraws != -1, CPL_ERROR_ILLEGAL_INPUT);

    nx = cpl_image_get_size_x(cpl_imagelist_get_const(rawbiases, 0));
    ny = cpl_image_get_size_y(cpl_imagelist_get_const(rawbiases, 0));
    cpl_ensure_code(nx != -1 && ny != -1, CPL_ERROR_ILLEGAL_INPUT);

    if(nx < detmon_ronbias_config.prescan_urx ||
       nx < detmon_ronbias_config.overscan_urx ||
       ny < detmon_ronbias_config.prescan_ury ||
       ny < detmon_ronbias_config.overscan_ury) {
        cpl_msg_warning(cpl_func, "PREOVERSCAN method not applied. Given "
                        "limits of prescan and overscan area "
                        "exceed image size. Please check and rerun.");
        return CPL_ERROR_NONE;
    }

    meanspre = cpl_vector_new(nraws);
    medspre = cpl_vector_new(nraws);
    rmsspre = cpl_vector_new(nraws);
    meansover = cpl_vector_new(nraws);
    medsover = cpl_vector_new(nraws);
    rmssover = cpl_vector_new(nraws);

    for(i = 0; i < nraws; i++) {
        double mean = 0;
        double stdev = 0;
        double median_prescan, median_overscan;

        cpl_image              *prescan = NULL;
        cpl_image              *overscan = NULL;

        const cpl_image        *c_raw = cpl_imagelist_get_const(rawbiases, i);

        cpl_ensure_code(c_raw != NULL, CPL_ERROR_ILLEGAL_INPUT);

        prescan =
            cpl_image_extract(c_raw,
                              detmon_ronbias_config.prescan_llx,
                              detmon_ronbias_config.prescan_lly,
                              detmon_ronbias_config.prescan_urx,
                              detmon_ronbias_config.prescan_ury);
        cpl_ensure_code(prescan != NULL, CPL_ERROR_ILLEGAL_INPUT);
        overscan =
            cpl_image_extract(c_raw,
                              detmon_ronbias_config.overscan_llx,
                              detmon_ronbias_config.overscan_lly,
                              detmon_ronbias_config.overscan_urx,
                              detmon_ronbias_config.overscan_ury);
        cpl_ensure_code(overscan != NULL, CPL_ERROR_ILLEGAL_INPUT);

        if(i == 0) {
            *synthetic = detmon_build_synthetic_from_pre_overscan(prescan,
                                                                  overscan);
            cpl_msg_info(cpl_func, "Creating SYNTHETIC frame");
            if(*synthetic == NULL) {
                cpl_msg_error(cpl_func, "Error creating SYNTHETIC frame");
                return CPL_ERROR_UNSPECIFIED;
            }
        }

        median_prescan  = cpl_image_get_median(prescan);
        median_overscan = cpl_image_get_median(overscan);
        cpl_image_delete(prescan);
        cpl_image_delete(overscan);

        error = irplib_ksigma_clip(c_raw,
                                   detmon_ronbias_config.
                                   prescan_llx,
                                   detmon_ronbias_config.
                                   prescan_lly,
                                   detmon_ronbias_config.
                                   prescan_urx,
                                   detmon_ronbias_config.
                                   prescan_ury,
                                   (double) detmon_ronbias_config.
                                   stacking_ks_low,
                                   detmon_ronbias_config.
                                   stacking_ks_iter, 1e-5,
                                   &mean, &stdev);
        cpl_ensure_code(!error, error);

        cpl_ensure_code(mean != 0 && stdev != 0, CPL_ERROR_UNSPECIFIED);

        error = cpl_vector_set(medspre, i, median_prescan);
        cpl_ensure_code(!error, error);

        error = cpl_vector_set(meanspre, i, mean);
        cpl_ensure_code(!error, error);
        error = cpl_vector_set(rmsspre, i, stdev);
        cpl_ensure_code(!error, error);
        error = irplib_ksigma_clip(c_raw,
                                   detmon_ronbias_config.
                                   overscan_llx,
                                   detmon_ronbias_config.
                                   overscan_lly,
                                   detmon_ronbias_config.
                                   overscan_urx,
                                   detmon_ronbias_config.
                                   overscan_ury,
                                   (double) detmon_ronbias_config.
                                   stacking_ks_low,
                                   detmon_ronbias_config.
                                   stacking_ks_iter, 1e-5,
                                   &mean, &stdev);
        cpl_ensure_code(!error, error);

        cpl_ensure_code(mean != 0 && stdev != 0, CPL_ERROR_UNSPECIFIED);

        error = cpl_vector_set(medsover, i, median_overscan);
        cpl_ensure_code(!error, error);

        error = cpl_vector_set(meansover, i, mean);
        cpl_ensure_code(!error, error);
        error = cpl_vector_set(rmssover, i, stdev);
        cpl_ensure_code(!error, error);

    }

    error = cpl_propertylist_append_double(qclist,DETMON_QC_BIAS_PRESCAN_MEAN,
                                           cpl_vector_get_mean(meanspre));

    error = cpl_propertylist_set_comment(qclist,DETMON_QC_BIAS_PRESCAN_MEAN,
                                         DETMON_QC_BIAS_PRESCAN_MEAN_C);

    cpl_ensure_code(!error, error);
    error = cpl_propertylist_append_double(qclist,DETMON_QC_BIAS_PRESCAN_MED,
                                           cpl_vector_get_mean(medspre));
    error = cpl_propertylist_set_comment(qclist,DETMON_QC_BIAS_PRESCAN_MED,
                                         DETMON_QC_BIAS_PRESCAN_MED_C);

    cpl_ensure_code(!error, error);
    error = cpl_propertylist_append_double(qclist,DETMON_QC_BIAS_PRESCAN_RON,
                                           cpl_vector_get_mean(rmsspre));

    error = cpl_propertylist_set_comment(qclist,DETMON_QC_BIAS_PRESCAN_RON,
                                         DETMON_QC_BIAS_PRESCAN_RON_C);
    cpl_ensure_code(!error, error);

    error =
        cpl_propertylist_append_double(qclist,DETMON_QC_BIAS_OVERSCAN_MEAN,
                                       cpl_vector_get_mean(meansover));
    error = cpl_propertylist_set_comment(qclist,DETMON_QC_BIAS_OVERSCAN_MEAN,
                                         DETMON_QC_BIAS_OVERSCAN_MEAN_C);
    cpl_ensure_code(!error, error);
    error = cpl_propertylist_append_double(qclist,DETMON_QC_BIAS_OVERSCAN_MED,
                                           cpl_vector_get_mean(medsover));
    error = cpl_propertylist_set_comment(qclist,DETMON_QC_BIAS_OVERSCAN_MED,
                                         DETMON_QC_BIAS_OVERSCAN_MED_C);
    cpl_ensure_code(!error, error);
    error = cpl_propertylist_append_double(qclist,DETMON_QC_BIAS_OVERSCAN_RON,
                                           cpl_vector_get_mean(rmssover));
    error = cpl_propertylist_set_comment(qclist,DETMON_QC_BIAS_OVERSCAN_RON,
                                         DETMON_QC_BIAS_OVERSCAN_RON_C);
    cpl_ensure_code(!error, error);

    /* The following seems not to be necessary.
       Pending of revision to be removed */
    /*
       error =
       cpl_propertylist_append_double(qclist,
       "ESO QC BIAS PRESCAN MEAN STDEV",
       cpl_vector_get_stdev(meanspre));
       cpl_ensure_code(!error, error);
       error =
       cpl_propertylist_append_double(qclist,
       "ESO QC BIAS PRESCAN MED STDEV",
       cpl_vector_get_stdev(medspre));
       cpl_ensure_code(!error, error);
       error =
       cpl_propertylist_append_double(qclist,
       "ESO QC BIAS PRESCAN RMS STDEV",
       cpl_vector_get_stdev(rmsspre));
       cpl_ensure_code(!error, error);

       error =
       cpl_propertylist_append_double(qclist,
       "ESO QC BIAS OVERSCAN MEAN STDEV",
       cpl_vector_get_stdev(meansover));
       cpl_ensure_code(!error, error);
       error =
       cpl_propertylist_append_double(qclist,
       "ESO QC BIAS OVERSCAN MED STDEV",
       cpl_vector_get_stdev(medsover));
       cpl_ensure_code(!error, error);
       error =
       cpl_propertylist_append_double(qclist,
       "ESO QC BIAS OVERSCAN RMS STDEV",
       cpl_vector_get_stdev(rmssover));
       cpl_ensure_code(!error, error);
       */

    cpl_vector_delete(meanspre);
    cpl_vector_delete(medspre);
    cpl_vector_delete(rmsspre);
    cpl_vector_delete(meansover);
    cpl_vector_delete(medsover);
    cpl_vector_delete(rmssover);

    return CPL_ERROR_NONE;
}

/*---------------------------------------------------------------------------*/

/*
 * @brief Get RON on bias frames and statistics on master bias on region for QC
 * @param rawbiases        Input image list
 * @param masterbias       Input master bias
 * @param qclist           input/output propertylist with QC parameters
 * @return CPL_ERROR_NONE on success.
 */

/*---------------------------------------------------------------------------*/
static cpl_error_code
detmon_compute_bias_stats_and_ron_in_region(const cpl_imagelist * rawbiases,
                      const cpl_image * masterbias,
                      cpl_propertylist * qclist)
{

    int                     nraws = cpl_imagelist_get_size(rawbiases);
    int                     i;

    int                     nx =
        cpl_image_get_size_x(cpl_imagelist_get_const(rawbiases, 0));
    int                     ny =
        cpl_image_get_size_y(cpl_imagelist_get_const(rawbiases, 0));

    cpl_vector *rmssreg;
    cpl_error_code          error;

    const cpl_image * c_raw;
    double median, mbias, mstdev;

    if(!strcmp(detmon_ronbias_config.pmethod, "DIF")) nraws--;

    rmssreg = cpl_vector_new(nraws);

    if(nx < detmon_ronbias_config.ref_urx ||
       ny < detmon_ronbias_config.ref_ury) {
        cpl_msg_warning(cpl_func, "REGION method not applied. Given "
                        "limits of prescan and overscan area "
                        "exceed image size. Please check and rerun.");
        return CPL_ERROR_NONE;
    }

    for(i = 0; i < nraws; i++) {
        double                  mean = 0;
        double                  stdev = 0;
        if(strcmp(detmon_ronbias_config.pmethod, "DIF")) {
            c_raw = cpl_imagelist_get_const(rawbiases, i);
        } else {
            const cpl_image        *c1_raw = cpl_imagelist_get_const(rawbiases, i);
            const cpl_image        * c2_raw = cpl_imagelist_get_const(rawbiases, i+1);
            c_raw = cpl_image_subtract_create(c1_raw, c2_raw);
        }
        error = irplib_ksigma_clip(c_raw,
                                   detmon_ronbias_config.ref_llx,
                                   detmon_ronbias_config.ref_lly,
                                   detmon_ronbias_config.ref_urx,
                                   detmon_ronbias_config.ref_ury,
                                   (double) detmon_ronbias_config.
                                   stacking_ks_low,
                                   detmon_ronbias_config.
                                   stacking_ks_iter, 1e-5,
                                   &mean, &stdev);
        cpl_ensure_code(!error, error);
        /*        cpl_vector_set(rmssreg, i, cpl_image_get_stdev_window(c_raw,
                  detmon_ronbias_config.ref_llx,
                  detmon_ronbias_config.ref_lly,
                  detmon_ronbias_config.ref_urx,
                  detmon_ronbias_config.ref_ury));
                  */
        error = cpl_vector_set(rmssreg, i, stdev);
        cpl_ensure_code(!error, error);
        if(!strcmp(detmon_ronbias_config.pmethod, "DIF")) cpl_image_delete((cpl_image *)c_raw);
    }

    median =  cpl_image_get_median_window(masterbias,
                                          detmon_ronbias_config.ref_llx,
                                          detmon_ronbias_config.ref_lly,
                                          detmon_ronbias_config.ref_urx,
                                          detmon_ronbias_config.ref_ury);
    error = irplib_ksigma_clip(masterbias,
                               detmon_ronbias_config.ref_llx,
                               detmon_ronbias_config.ref_lly,
                               detmon_ronbias_config.ref_urx,
                               detmon_ronbias_config.ref_ury,
                               (double) detmon_ronbias_config.
                               stacking_ks_low,
                               detmon_ronbias_config.
                               stacking_ks_iter, 1e-5,
                               &mbias, &mstdev);

    error = cpl_propertylist_append_double(qclist,DETMON_QC_BIAS_REGION_MED,
                                           median);
    error = cpl_propertylist_set_comment(qclist,DETMON_QC_BIAS_REGION_MED,
                                         DETMON_QC_BIAS_REGION_MED_C);
    cpl_ensure_code(!error, error);

    error = cpl_propertylist_append_double(qclist,DETMON_QC_BIAS_REGION_VAL,
                                           mbias);
    error = cpl_propertylist_set_comment(qclist,DETMON_QC_BIAS_REGION_VAL,
                                         DETMON_QC_BIAS_REGION_VAL_C);
    cpl_ensure_code(!error, error);
    error = cpl_propertylist_append_double(qclist,DETMON_QC_BIAS_REGION_RON,
                                           cpl_vector_get_mean(rmssreg));
    error = cpl_propertylist_set_comment(qclist,DETMON_QC_BIAS_REGION_RON,
                                         DETMON_QC_BIAS_REGION_RON_C);
    cpl_ensure_code(!error, error);
    /*
       error =
       cpl_propertylist_append_double(qclist, "ESO QC BIAS REGION RMS STDEV",
       cpl_vector_get_stdev(rmssreg));
       cpl_ensure_code(!error, error);
       */
    cpl_vector_delete(rmssreg);

    return cpl_error_get_code();
}

/*---------------------------------------------------------------------------*/

/*
 * @brief  Determines master bias and QC parameters
 * @param  rawbiases        Input raw images
 * @param  bpmhot           hot pixel bpm
 * @param  bpmcold          cold pixel bpm
 * @param  bpmdev           deviant pixel bpm
 * @param  qclist           fits header
 * @return master bias on success or NULL
 */

/*---------------------------------------------------------------------------*/
static cpl_image *
detmon_ronbias_master(const cpl_imagelist * rawbiases,
                      cpl_mask ** bpmhot, cpl_mask ** bpmcold,
                      cpl_mask ** bpmdev, cpl_propertylist * qclist)
{
    double                  mean = 0;
    double                  stdev = 0;
    cpl_image              *masterbias = NULL;
    double dark_med, stdev_med,lower, upper;
    int hotpix_nb, coldpix_nb, devpix_nb;
    cpl_image * stdev_im = NULL;

    if(!strcmp(detmon_ronbias_config.stacking_method, "MEAN"))
        masterbias = cpl_imagelist_collapse_create(rawbiases);
    if(!strcmp(detmon_ronbias_config.stacking_method, "MINMAX"))
        masterbias =
            cpl_imagelist_collapse_minmax_create(rawbiases, 0, 10000);
    if(!strcmp(detmon_ronbias_config.stacking_method, "KSIGMA"))
        masterbias =
            cpl_imagelist_collapse_sigclip_create(rawbiases, 3.0, 3.0, 0.9,
                                                  CPL_COLLAPSE_MEAN, NULL);
    if(!strcmp(detmon_ronbias_config.stacking_method, "MEDIAN"))
        masterbias = cpl_imagelist_collapse_median_create(rawbiases);

    skip_if(masterbias == NULL);

    skip_if(irplib_ksigma_clip(masterbias, 1, 1,
                               cpl_image_get_size_x(masterbias),
                               cpl_image_get_size_y(masterbias),
                               (double) detmon_ronbias_config.
                               stacking_ks_low,
                               detmon_ronbias_config.
                               stacking_ks_iter, 1e-5,
                               &mean, &stdev));

    if(irplib_isnan(mean))
        cpl_msg_error(cpl_func, "We have an error in mean");
    if(irplib_isnan(stdev))
        cpl_msg_error(cpl_func, "We have an error in stdev");

    skip_if(cpl_propertylist_append_double(qclist,DETMON_QC_MASTER_MEAN,
                                           mean));
    skip_if(cpl_propertylist_set_comment(qclist,DETMON_QC_MASTER_MEAN,
                                         DETMON_QC_MASTER_MEAN_C));

    skip_if(cpl_propertylist_append_double(qclist,DETMON_QC_MASTER_RMS,
                                           stdev));
    skip_if(cpl_propertylist_set_comment(qclist,DETMON_QC_MASTER_RMS,
                                         DETMON_QC_MASTER_RMS_C));

    /* Compute median-rms of the central part of the dark  */
    dark_med = cpl_image_get_median(masterbias);

    lower = dark_med - stdev * detmon_ronbias_config.stacking_ks_low;
    upper = dark_med + stdev * detmon_ronbias_config.stacking_ks_high;

    /* Create the hot pixel map */
    cpl_mask_delete(*bpmhot);
    irplib_check(*bpmhot = cpl_mask_threshold_image_create(masterbias,
                                                           upper, DBL_MAX),
                 "Cannot compute the hot pixel map");
    hotpix_nb = cpl_mask_count(*bpmhot);
    skip_if (0);

    /* Create the cold pixel map */
    cpl_mask_delete(*bpmcold);
    irplib_check(*bpmcold = cpl_mask_threshold_image_create(masterbias,
                                                            -FLT_MAX, lower),
                 "Cannot compute the cold pixel map");
    coldpix_nb = cpl_mask_count(*bpmcold);
    skip_if (0);

    /* Create the deviant pixel map */
    stdev_im  = irplib_imagelist_collapse_stdev_create(rawbiases);
    stdev_med = cpl_image_get_median(stdev_im);

    skip_if(irplib_ksigma_clip(stdev_im, 1, 1,
                               cpl_image_get_size_x(stdev_im),
                               cpl_image_get_size_y(stdev_im),
                               (double) detmon_ronbias_config.
                               stacking_ks_low,
                               detmon_ronbias_config.
                               stacking_ks_iter, 1e-5,
                               &mean, &stdev));

    lower = stdev_med - stdev * detmon_ronbias_config.stacking_ks_low;
    upper = stdev_med + stdev * detmon_ronbias_config.stacking_ks_high;

    cpl_mask_delete(*bpmdev);
    irplib_check(*bpmdev = cpl_mask_threshold_image_create(stdev_im,
                                                           lower, upper),
                 "Cannot compute the cold pixel map");
    cpl_mask_not(*bpmdev);
    devpix_nb = cpl_mask_count(*bpmdev);
    skip_if (0);


    skip_if(cpl_propertylist_append_int(qclist,DETMON_QC_NBCOLDPIX,coldpix_nb));
    skip_if(cpl_propertylist_set_comment(qclist,DETMON_QC_NBCOLDPIX,
                                         DETMON_QC_NBCOLDPIX_C));

    skip_if(cpl_propertylist_append_int(qclist,DETMON_QC_NBHOTPIX, hotpix_nb));
    skip_if(cpl_propertylist_set_comment(qclist,DETMON_QC_NBHOTPIX,
                                         DETMON_QC_NBHOTPIX_C));

    skip_if(cpl_propertylist_append_int(qclist,DETMON_QC_NBDEVPIX, devpix_nb));
    skip_if(cpl_propertylist_set_comment(qclist,DETMON_QC_NBDEVPIX,
                                         DETMON_QC_NBDEVPIX_C));

    end_skip;

    cpl_image_delete(stdev_im);

    if (cpl_error_get_code()) {
        cpl_image_delete(masterbias);
        masterbias = NULL;
    }

    return masterbias;
}

/*---------------------------------------------------------------------------*/

/*
 * @brief  save a ronbias results on a FITS file
 * @param  parlist        input parameters
 * @param  frameset       input frames
 * @param  recipe_name    input recipe name
 * @param  pipeline_name  input pipeline id
 * @param  pafregexp      regular expression to remove from FITS header
 * @param  pro_master     pro catg master frame
 * @param  pro_xstr       pro catg xstructure
 * @param  pro_ystr       pro catg ystructure
 * @param  pro_synth      pro catg synthetic frame
 * @param  pro_bpmhot     pro catg hot pix BPM
 * @param  pro_pmpcold    pro catg cold pix BPM
 * @param  pro_ppmdev     pro catg deviant pix BPM
 * @param  package        package name
 * @param  masterbias     master bias
 * @param  synthetic      synthetic frame
 * @param  bpmhot         hot pixel bpm
 * @param  bpmcold        cold pixel bpm
 * @param  bpmdev         deviant pixel bpm
 * @param  qclist         qc parameter (header)
 * @param  flag_sets      switch
 * @param  which_set      id to specify a given set
 * @param  usedframes     used frames
 * @param  whichset       id to specify a given set
 * @return CPL_ERROR_NONE on success.
 */

/*---------------------------------------------------------------------------*/
static                  cpl_error_code
detmon_ronbias_save(const cpl_parameterlist * parlist,
                    cpl_frameset * frameset,
                    const char *recipe_name,
                    const char *pipeline_name,
                    const char *pafregexp,
                    const cpl_propertylist * pro_master,
                    const cpl_propertylist * pro_xstr, /* Unsupported*/
                    const cpl_propertylist * pro_ystr, /* Unsupported*/
                    const cpl_propertylist * pro_synth,
                    const cpl_propertylist * pro_bpmhot,
                    const cpl_propertylist * pro_bpmcold,
                    const cpl_propertylist * pro_bpmdev,
                    const char *package,
                    const cpl_image * masterbias,
                    const cpl_image * synthetic,
                    const cpl_mask * bpmhot,
                    const cpl_mask * bpmcold,
                    const cpl_mask * bpmdev,
                    cpl_propertylist * qclist,
                    const int flag_sets,
                    const int which_set,
                    cpl_frameset * usedframes,
                    int whichext)
{

    cpl_frame              *ref_frame;
    cpl_propertylist       *plist = NULL;
    char                   *name_o = NULL; /* Avoid (false) uninit warning */

    cpl_propertylist       * paflist   = NULL;
    cpl_propertylist       * mainplist = NULL;
    cpl_propertylist       * xplist    = NULL;
    cpl_image              * image     = NULL;

    cpl_propertylist  * mypro_master     =
        cpl_propertylist_duplicate(pro_master);

    cpl_propertylist  * mypro_synth = NULL;
    cpl_propertylist  * mypro_bpmhot    =
        cpl_propertylist_duplicate(pro_bpmhot);
    cpl_propertylist  * mypro_bpmcold =
        cpl_propertylist_duplicate(pro_bpmcold);
    cpl_propertylist  * mypro_bpmdev =
        cpl_propertylist_duplicate(pro_bpmdev);

    cpl_ensure_code(parlist != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(frameset != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(pafregexp != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(package != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(recipe_name != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(pipeline_name != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(usedframes != NULL, CPL_ERROR_NULL_INPUT);

    if (pro_synth)
        mypro_synth = cpl_propertylist_duplicate(pro_synth);

    /* Extra check while XSTR and YSTR are not supported */
    cpl_ensure_code(pro_xstr == NULL && pro_ystr == NULL,
                    CPL_ERROR_UNSUPPORTED_MODE);

    /* Extract extension headers if multi-extension */
    if (detmon_ronbias_config.exts < 0) {
        const char * filename =
            cpl_frame_get_filename(cpl_frameset_get_position(frameset, 0));


        xplist = cpl_propertylist_load_regexp(filename, whichext,
                                              "ESO DET", 0);
        skip_if(cpl_propertylist_append(xplist, qclist));
    }

    cpl_msg_info(cpl_func,"dealing with extention %d",whichext);

    /* This is only used later for PAF */
    /* Get FITS header from reference file */
    ref_frame = cpl_frameset_get_position(frameset, 0);
    skip_if(ref_frame == NULL);

    skip_if((mainplist =
             cpl_propertylist_load(cpl_frame_get_filename(ref_frame),
                                   0)) == NULL);

    /**************************/
    /*  Write the MASTERBIAS  */
    /**************************/

    /* Set the file name for the table */
    if(!flag_sets) {
        name_o = cpl_sprintf("%s_masterbias.fits", recipe_name);
        assert(name_o != NULL);
    } else {
        name_o =
            cpl_sprintf("%s_masterbias_set%02d.fits", recipe_name,
                        which_set);
        assert(name_o != NULL);
    }
    /* Save the MASTERBIAS image */
    if (whichext == 0) {
        cpl_propertylist_append(mypro_master, qclist);

        skip_if(cpl_dfs_save_image(frameset, NULL, parlist, usedframes, NULL,
                                   masterbias, CPL_BPP_IEEE_FLOAT, recipe_name,
                                   mypro_master, NULL, package, name_o));
    } else
        skip_if(cpl_image_save(masterbias,
                               name_o, CPL_BPP_IEEE_FLOAT, xplist,
                               CPL_IO_EXTEND));

    /* Free */
    cpl_free(name_o);
    name_o = NULL;

    /*****************************/
    /*  Write the HOT PIXEL MAP  */
    /*****************************/

    /* Set the file name for the table */
    if(!flag_sets) {
        name_o = cpl_sprintf("%s_hotpixmap.fits", recipe_name);
        assert(name_o != NULL);
    } else {
        name_o =
            cpl_sprintf("%s_hotpixmap_set%02d.fits", recipe_name,
                        which_set);
        assert(name_o != NULL);
    }
    /* Save the HOTBPM image */
    skip_if(0);
    image = cpl_image_new_from_mask(bpmhot);
    cpl_error_reset();

    if (whichext == 0) {
        cpl_propertylist_append(mypro_bpmhot, qclist);

        skip_if(cpl_dfs_save_image(frameset, NULL, parlist, usedframes, NULL,
                                   image, CPL_BPP_IEEE_FLOAT, recipe_name,
                                   mypro_bpmhot, NULL, package, name_o));
    } else
        skip_if(cpl_image_save(image,
                               name_o, CPL_BPP_IEEE_FLOAT, xplist,
                               CPL_IO_EXTEND));

    /* Free */
    cpl_free(name_o);
    cpl_image_delete(image);
    image = NULL;
    name_o = NULL;

    /*****************************/
    /*  Write the COLD PIXEL MAP  */
    /*****************************/

    /* Set the file name for the table */
    if(!flag_sets) {
        name_o = cpl_sprintf("%s_coldpixmap.fits", recipe_name);
        assert(name_o != NULL);
    } else {
        name_o =
            cpl_sprintf("%s_coldpixmap_set%02d.fits", recipe_name,
                        which_set);
        assert(name_o != NULL);
    }
    /* Save the COLDBPM image */
    skip_if(0);
    image = cpl_image_new_from_mask(bpmcold);
    cpl_error_reset();

    if (whichext == 0) {
        cpl_propertylist_append(mypro_bpmcold, qclist);

        skip_if(cpl_dfs_save_image(frameset, NULL, parlist, usedframes, NULL,
                                   image, CPL_BPP_IEEE_FLOAT, recipe_name,
                                   mypro_bpmcold, NULL, package, name_o));
    } else
        skip_if(cpl_image_save(image,
                               name_o, CPL_BPP_IEEE_FLOAT, xplist,
                               CPL_IO_EXTEND));

    /* Free */
    cpl_free(name_o);
    cpl_image_delete(image);
    image = NULL;
    name_o = NULL;

    /*****************************/
    /*  Write the DEV PIXEL MAP  */
    /*****************************/

    /* Set the file name for the table */
    if(!flag_sets) {
        name_o = cpl_sprintf("%s_devpixmap.fits", recipe_name);
        assert(name_o != NULL);
    } else {
        name_o =
            cpl_sprintf("%s_devpixmap_set%02d.fits", recipe_name,
                        which_set);
        assert(name_o != NULL);
    }
    /* Save the DEVBPM image */
    skip_if(0);
    image = cpl_image_new_from_mask(bpmdev);
    cpl_error_reset();

    if (whichext == 0) {
        cpl_propertylist_append(mypro_bpmdev, qclist);

        skip_if(cpl_dfs_save_image(frameset, NULL,parlist, usedframes, NULL,
                                   image, CPL_BPP_IEEE_FLOAT, recipe_name,
                                   mypro_bpmdev, NULL, package, name_o));
    } else
        skip_if(cpl_image_save(image,
                               name_o, CPL_BPP_IEEE_FLOAT, xplist,
                               CPL_IO_EXTEND));

    /* Free */
    cpl_free(name_o);
    cpl_image_delete(image);
    image = NULL;
    name_o = NULL;

    /*******************************/
    /*  Write the SYNTHETIC        */
    /*******************************/
    if(detmon_ronbias_config.method_bitmask & PREOVERSCAN) {
        /* Set the file name for the table */
        if(!flag_sets) {
            name_o = cpl_sprintf("%s_synthetic.fits", recipe_name);
            assert(name_o != NULL);
        } else {
            name_o =
                cpl_sprintf("%s_synthetic_set%02d.fits", recipe_name,
                            which_set);
            assert(name_o != NULL);
        }

        if (whichext == 0) {
            /* Save the SYNTHETIC image */
            cpl_propertylist_append(mypro_synth, qclist);

            skip_if(cpl_dfs_save_image(frameset, NULL, parlist, usedframes,
                                       NULL,synthetic, CPL_BPP_IEEE_DOUBLE,
                                       recipe_name, mypro_synth, NULL,
                                       package, name_o));
        } else
            skip_if(cpl_image_save(synthetic, name_o, CPL_BPP_IEEE_FLOAT,
                                   xplist, CPL_IO_EXTEND));

        /* Free */
        cpl_free(name_o);
        name_o = NULL;
    }

    /*******************************/
    /*  Write the PAF file         */
    /*******************************/
    if (qclist) {
        paflist = cpl_propertylist_new();

        /* Set the file name for the PAF */
        if(detmon_ronbias_config.exts >= 0) {
            skip_if((plist =
                     cpl_propertylist_load(cpl_frame_get_filename(ref_frame),
                                           detmon_ronbias_config.exts)) == NULL);

            if(!flag_sets) {
                name_o = cpl_sprintf("%s.paf", recipe_name);
                assert(name_o != NULL);
            } else {
                name_o = cpl_sprintf("%s_set%02d.paf",
                                     recipe_name, which_set);
                assert(name_o != NULL);
            }
        } else {
            skip_if((plist =
                     cpl_propertylist_load(cpl_frame_get_filename(ref_frame),
                                           whichext)) == NULL);


            if(!flag_sets) {
                name_o = cpl_sprintf("%s_ext%02d.paf",
                                     recipe_name, whichext);
                assert(name_o != NULL);
            } else {
                name_o = cpl_sprintf("%s_set%02d_ext%02d.paf",
                                     recipe_name,
                                     which_set, whichext);
                assert(name_o != NULL);
            }
        }

        /* Get the keywords for the paf file */
        skip_if(cpl_propertylist_copy_property_regexp(paflist, plist,
                                                      pafregexp, 0));
        skip_if(cpl_propertylist_copy_property_regexp(paflist, mainplist,
                                                      pafregexp, 0));

        skip_if(cpl_propertylist_append(paflist, qclist));

        /* Save the PAF */
        skip_if(cpl_dfs_save_paf(pipeline_name, recipe_name, paflist, name_o));

    }

    end_skip;

    cpl_propertylist_delete(plist);
    cpl_propertylist_delete(paflist);
    cpl_propertylist_delete(mainplist);
    cpl_propertylist_delete(xplist);
    cpl_free(name_o);
    cpl_image_delete(image);

    cpl_propertylist_delete(mypro_master);
    cpl_propertylist_delete(mypro_synth);
    cpl_propertylist_delete(mypro_bpmhot);
    cpl_propertylist_delete(mypro_bpmcold);
    cpl_propertylist_delete(mypro_bpmdev);

    return cpl_error_get_code();
}

/*---------------------------------------------------------------------------*/
/*
 * @brief  Classify input data with a given tag
 * @param  set    Input frameset
 * @param  tag    Input tag
 * @return 0 on success, -1 if failure
 */
/*---------------------------------------------------------------------------*/
static int
detmon_ronbias_dfs_set_groups(cpl_frameset * set, const char *tag)
{


    int                     nframes;
    int                     i;

    /* Check entries */
    if(set == NULL)
        return -1;

    /* Initialize */
    nframes = cpl_frameset_get_size(set);

    /* Loop on frames */
    for(i = 0; i < nframes; i++) {
        cpl_frame* cur_frame = cpl_frameset_get_position(set, i);
        const char* cur_tag = cpl_frame_get_tag(cur_frame);

        /* RAW frames */
        if(!strcmp(cur_tag, tag))
            cpl_frame_set_group(cur_frame, CPL_FRAME_GROUP_RAW);
        /* CALIB frames */

        /*        else if (!strcmp(tag, IIINSTRUMENT_CALIB_FLAT))
                  cpl_frame_set_group(cur_frame, CPL_FRAME_GROUP_CALIB) ;
                  */
    }
    return 0;
}

/*---------------------------------------------------------------------------*/

/*
 * @brief  build synthetic frame from pre-overscan images
 *
 * @param  prescan        prescan image
 * @param  overscan       overscan image
 *
 * @return CPL_ERROR_NONE on success.
 */

/*---------------------------------------------------------------------------*/
static cpl_image *
detmon_build_synthetic_from_pre_overscan(const cpl_image * prescan,
                                         const cpl_image * overscan)
{
    cpl_size                     j;

    int distance = detmon_ronbias_config.overscan_urx -
        detmon_ronbias_config.prescan_llx + 1;

    double * mean_x = (double *) cpl_malloc(sizeof(double) * distance);

    double * xvalues = (double *) cpl_malloc(sizeof(double) * distance);

    cpl_vector             *x = NULL;
    cpl_vector             *y = NULL;

    cpl_polynomial         *poly = NULL;
    cpl_polynomial         *poly2 = NULL;

    cpl_matrix     * samppos;

    cpl_size         pows[2] = { 0, 0 };

    cpl_image      * synthetic = NULL;

    double           initial = 0;

    /* Initialize */
    for(j = 0; j < distance; j++) {
        *(mean_x + j) = 0;
        *(xvalues + j) = j;
    }

    for(j = 0; j < cpl_image_get_size_x(prescan); j++) {
        *(mean_x + j) =
            cpl_image_get_mean_window(prescan, j + 1, 1, j + 1,
                                      cpl_image_get_size_y(prescan));
    }

    for(j = 0; j < cpl_image_get_size_x(overscan); j++) {
        *(mean_x + distance - cpl_image_get_size_x(overscan) + j) =
            cpl_image_get_mean_window(overscan, j + 1, 1, j + 1,
                                      cpl_image_get_size_y(overscan));
    }

    x = cpl_vector_wrap(distance, xvalues);
    y = cpl_vector_wrap(distance, mean_x);

    poly = cpl_polynomial_new(1);
    samppos =
        cpl_matrix_wrap(1, cpl_vector_get_size(x), cpl_vector_get_data(x));

    cpl_polynomial_fit(poly, samppos, NULL, y, NULL,
                       CPL_FALSE, NULL, &detmon_ronbias_config.preoverscan_degree);

    cpl_matrix_unwrap(samppos);

    cpl_vector_unwrap(x);
    cpl_vector_unwrap(y);

    initial = *mean_x;

    cpl_free(xvalues);
    cpl_free(mean_x);

    poly2 = cpl_polynomial_new(2);

    j = 0;
    cpl_polynomial_set_coeff(poly2, pows, cpl_polynomial_get_coeff(poly, &j));

    pows[0] = 1;
    j = 1;
    cpl_polynomial_set_coeff(poly2, pows, cpl_polynomial_get_coeff(poly, &j));

    cpl_polynomial_delete(poly);

    synthetic =
        cpl_image_new(distance, cpl_image_get_size_y(prescan),
                      CPL_TYPE_DOUBLE);

    if(cpl_image_fill_polynomial(synthetic, poly2, initial, 1, 1, 1)) {
        cpl_msg_error(cpl_func, "Error creating the synthetic frame");
        cpl_polynomial_delete(poly2);
        return NULL;
    }

    cpl_polynomial_delete(poly2);

    return synthetic;
}

/*---------------------------------------------------------------------------*/

/*
 * @brief  computes dutycycle
 * @param  frameset       Input frameset
 * @param  qclist         header to hold qc parameter
 * @return CPL_ERROR_NONE on success.
 */

/*---------------------------------------------------------------------------*/
static cpl_error_code
detmon_ronbias_dutycycl(const cpl_frameset * frameset,
                        cpl_propertylist * qclist)
{
    const cpl_frame        *first = 0;
    cpl_propertylist       *plistfirst = 0;
    double                  tfirst;
    int                     nraws;
    const cpl_frame        *last = 0;
    cpl_propertylist       *plistlast = 0;
    double                  tlast;
    double                  dutycycl;
    cpl_error_code          error;

    first = cpl_frameset_get_position_const(frameset, 0);
    plistfirst = cpl_propertylist_load(cpl_frame_get_filename(first), 0);
    tfirst = cpl_propertylist_get_double(plistfirst, "MJD-OBS");
    nraws = cpl_frameset_get_size(frameset);
    last = cpl_frameset_get_position_const(frameset, nraws - 1);
    plistlast = cpl_propertylist_load(cpl_frame_get_filename(last), 0);
    tlast = cpl_propertylist_get_double(plistlast, "MJD-OBS");
    dutycycl = (tlast - tfirst) / (nraws - 1);

    error = cpl_error_get_code();
    if (error != CPL_ERROR_NONE)
    {
        goto cleanup;
    }
    cpl_propertylist_append_double(qclist,DETMON_QC_DUTYCYCL, dutycycl);
    error = cpl_propertylist_set_comment(qclist,DETMON_QC_DUTYCYCL,
                                         DETMON_QC_DUTYCYCL_C);

cleanup:

    cpl_propertylist_delete(plistfirst);
    cpl_propertylist_delete(plistlast);

    return error;
}


/* Start duplicated code */

#define RECT_RON_HS         4
#define RECT_RON_SAMPLES    100

/*---------------------------------------------------------------------------*/

/*
 * @brief  Computes bias level and ron in a region
 * @param  diff        Input difference image
 * @param  zone_def    Input zone definition
 * @param  ron_hsize   window half size to compute ron
 * @param  ron_nsamp   number of (randomly distributed) sampling windows
 * @param  bias        output bias level
 * @param  error       output error associated to bias level
 * @return CPL_ERROR_NONE on success.
 */

/*---------------------------------------------------------------------------*/
cpl_error_code
irplib_flux_get_bias_window(const cpl_image * diff,
                            const int *zone_def,
                            int ron_hsize,
                            int ron_nsamp, double *bias, double *error)
{
    const int               hsize = ron_hsize < 0 ? RECT_RON_HS : ron_hsize;
    const int               nsamples =
        ron_nsamp < 0 ? RECT_RON_SAMPLES : ron_nsamp;
    cpl_bivector           *sample_reg;
    cpl_vector             *rms_list;
    int                     rect[4];
    int                     zone[4];
    double                 *px;
    double                 *py;
    double                 *pr;
    int                     i;

    /* Test entries */
    cpl_ensure_code(diff && bias, CPL_ERROR_NULL_INPUT);

    /* Generate nsamples window centers in the image */
    if(zone_def != NULL) {
        rect[0] = zone_def[0] + hsize + 1;      /* xmin */
        rect[1] = zone_def[1] - hsize - 1;      /* xmax */
        rect[2] = zone_def[2] + hsize + 1;      /* ymin */
        rect[3] = zone_def[3] - hsize - 1;      /* ymax */
    } else {
        rect[0] = hsize + 1;    /* xmin */
        rect[1] = cpl_image_get_size_x(diff) - hsize - 1;       /* xmax */
        rect[2] = hsize + 1;    /* ymin */
        rect[3] = cpl_image_get_size_y(diff) - hsize - 1;       /* ymax */
    }

    cpl_ensure_code(rect[0] < rect[1] && rect[2] < rect[3],
                    CPL_ERROR_ILLEGAL_INPUT);

    /* Generate n+1 regions, because the first region is always at (0,0) */
    /* and it would bias the measurement. */
    sample_reg =
        irplib_bivector_gen_rect_poisson(rect, nsamples + 1, nsamples + 1);
    cpl_ensure(sample_reg != NULL, CPL_ERROR_ILLEGAL_INPUT,
               CPL_ERROR_ILLEGAL_INPUT);

    px = cpl_bivector_get_x_data(sample_reg);
    py = cpl_bivector_get_y_data(sample_reg);

    /* Now, for each window center, extract a vignette and compute the */
    /* signal RMS in it. Store this rms into a table. */
    rms_list = cpl_vector_new(nsamples);
    cpl_ensure(rms_list != NULL, CPL_ERROR_NULL_INPUT, CPL_ERROR_NULL_INPUT);
    pr = cpl_vector_get_data(rms_list);

    for(i = 0; i < nsamples; i++) {
        zone[0] = (int) px[i + 1] - hsize;
        zone[1] = (int) px[i + 1] + hsize;
        zone[2] = (int) py[i + 1] - hsize;
        zone[3] = (int) py[i + 1] + hsize;
        pr[i] = cpl_image_get_mean_window(diff,
                                          zone[0], zone[2], zone[1], zone[3]);
    }
    cpl_bivector_delete(sample_reg);

    /* The error is the rms of the rms */
    if(error != NULL)
        *error = cpl_vector_get_stdev(rms_list);

    /* The final computed RMS is the median of all values.  */
    /* This call will modify the rms_list */
    *bias = cpl_vector_get_median(rms_list);

    cpl_vector_delete(rms_list);

    return CPL_ERROR_NONE;
}

#undef RECT_RON_HS
#undef RECT_RON_SAMPLES

/*---------------------------------------------------------------------------*/

/*
 * @brief  Generates Poisson distributed values
 * @param  r        Input values
 * @param  np       number of points
 * @param  homog    homogeneity factor
 * @return CPL_ERROR_NONE on success.
 */

/*---------------------------------------------------------------------------*/
static cpl_bivector    *
irplib_bivector_gen_rect_poisson(const int *r, const int np, const int homog)
{
    double                  min_dist;
    int                     i;
    int                     gnp;
    cpl_bivector           *list;
    double                  cand_x, cand_y;
    int                     ok;
    int                     start_ndx;
    int                     xmin, xmax, ymin, ymax;

    /* Corrected Homogeneity factor */
    const int               homogc = 0 < homog && homog < np ? homog : np;
    double                 *px;
    double                 *py;

    /* error handling: test arguments are correct */
    cpl_ensure(r, CPL_ERROR_NULL_INPUT, NULL);
    cpl_ensure(np > 0, CPL_ERROR_ILLEGAL_INPUT, NULL);

    list = cpl_bivector_new(np);
    cpl_ensure(list, CPL_ERROR_NULL_INPUT, NULL);
    px = cpl_bivector_get_x_data(list);
    py = cpl_bivector_get_y_data(list);

    xmin = r[0];
    xmax = r[1];
    ymin = r[2];
    ymax = r[3];

    min_dist =
        CPL_MATH_SQRT1_2 * ((xmax - xmin) * (ymax - ymin) / (double) (homogc + 1));
    gnp = 1;
    px[0] = 0;
    py[0] = 0;

    /* First: generate <homog> points */
    while(gnp < homogc) {
        /* Pick a random point within requested range */
        cand_x = cpl_drand() * (xmax - xmin) + xmin;
        cand_y = cpl_drand() * (ymax - ymin) + ymin;

        /* Check the candidate obeys the minimal Poisson distance */
        ok = 1;
        for(i = 0; i < gnp; i++) {
            if(pdist(cand_x, cand_y, px[i], py[i]) < min_dist) {
                /* does not check Poisson law: reject point */
                ok = 0;
                break;
            }
        }
        if(ok) {
            /* obeys Poisson law: register the point as valid */
            px[gnp] = cand_x;
            py[gnp] = cand_y;
            gnp++;
        }
    }

    /* Iterative process: */
    /* Pick points out of Poisson distance of the last <homogc-1> points. */
    start_ndx = 0;
    while(gnp < np) {
        /* Pick a random point within requested range */
        cand_x = cpl_drand() * (xmax - xmin) + xmin;
        cand_y = cpl_drand() * (ymax - ymin) + ymin;

        /* Check the candidate obeys the minimal Poisson distance */
        ok = 1;
        for(i = 0; i < homogc; i++) {
            if(pdist(cand_x,
                     cand_y,
                     px[start_ndx + i], py[start_ndx + i]) < min_dist) {
                /* does not check Poisson law: reject point */
                ok = 0;
                break;
            }
        }
        if(ok) {
            /* obeys Poisson law: register the point as valid */
            px[gnp] = cand_x;
            py[gnp] = cand_y;
            gnp++;
        }
    }

    /* Iterative process: */
    /* Pick points out of Poisson distance of the last <homogc-1> points. */
    start_ndx = 0;
    while(gnp < np) {
        /* Pick a random point within requested range */
        cand_x = cpl_drand() * (xmax - xmin) + xmin;
        cand_y = cpl_drand() * (ymax - ymin) + ymin;

        /* Check the candidate obeys the minimal Poisson distance */
        ok = 1;
        for(i = 0; i < homogc; i++) {
            if(pdist(cand_x,
                     cand_y,
                     px[start_ndx + i], py[start_ndx + i]) < min_dist) {
                /* does not check Poisson law: reject point */
                ok = 0;
                break;
            }
        }
        if(ok) {
            /* obeys Poisson law: register the point as valid */
            px[gnp] = cand_x;
            py[gnp] = cand_y;
            gnp++;
            start_ndx++;
        }
    }
    return list;
}

/* End of duplicated code */


/*---------------------------------------------------------------------------*/
/**
  @brief    Collapse an imagelist to the stdev of each pixel position
  @param    imlist    the input images list
  @return   the average image or NULL on error case.
  @see cpl_imagelist_is_uniform()

  The returned image has to be deallocated with cpl_image_delete().

  The bad pixel maps of the images in the input list are taken into
  account, the result image pixels are flagged as rejected for those
  where there were no good pixel at the same position in the input image list.

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_ILLEGAL_INPUT if the input image list is not valid
 */
/*---------------------------------------------------------------------------*/
cpl_image *
irplib_imagelist_collapse_stdev_create(const cpl_imagelist * imlist)
{






    /* Check inputs */
    cpl_ensure(imlist != NULL, CPL_ERROR_NULL_INPUT, NULL);
    cpl_ensure(cpl_imagelist_is_uniform(imlist) == 0, CPL_ERROR_ILLEGAL_INPUT,
               NULL);

    /* Create mean image with its first iterative value = first image */
    cpl_image* mean = cpl_image_duplicate(cpl_imagelist_get_const(imlist, 0));
    cpl_image_fill_rejected(mean, 0.0);
    cpl_image_accept_all(mean);

    cpl_image* stdev = cpl_image_new(cpl_image_get_size_x(mean),
                          cpl_image_get_size_y(mean),
                          CPL_TYPE_FLOAT);

    for (int i = 1; i < cpl_imagelist_get_size(imlist); i++) {
        cpl_image* delta = cpl_image_subtract_create(cpl_imagelist_get_const(imlist, i),
                                          mean);
        cpl_image_fill_rejected(delta, 0.0);
        cpl_image_accept_all(delta);

        cpl_image* sq_delta = cpl_image_multiply_create(delta, delta);

        cpl_image_multiply_scalar(sq_delta, ((double) i / (double)(i+1)));
        cpl_image_add(stdev, sq_delta);

        cpl_image_divide_scalar(delta, i + 1);
        cpl_image_add(mean, delta);

        cpl_image_delete(delta);
        cpl_image_delete(sq_delta);
    }

    cpl_image_divide_scalar(stdev, cpl_imagelist_get_size(imlist) - 1);
    cpl_image_power(stdev, 0.5);

    cpl_image_delete(mean);

    return stdev;
}
/**@}*/
